/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

/* Algorithme des k moyennes.

   Les données sont ici des tableaux de `double` de dimension `d`.
   (La valeur de `d` est passée en paramètre aux fonctions.)
*/

#include "prelude.h"
#include "kmeans.h"

double kmeans_sqr_dist(int d, double *x, double *y) {
  double dist = 0.;
  for (int i = 0; i < d; i++)
    dist += (x[i] - y[i]) * (x[i] - y[i]);
  return dist;
}

// on initialise les moyennes par un tirage aléatoire dans les données
void kmeans_sampling(int k, int d, double **data, int n, double **mea) {
  assert(n >= k);
  for (int i = 0; i < k; i++)
    memcpy(mea[i], data[i], d * sizeof(double));
  for (int i = k; i < n; i++) {
    int j = rand() % (i+1);
    if (j < k)
      memcpy(mea[j], data[i], d * sizeof(double));
  }
}

// classifie `di` selon les moyennes dans `mea`
// écrit le résultat dans `*ci`
// renvoie true si la classification a changé
bool kmeans_classify(int k, int d, double *di, double **mea, int *cli) {
  int m = 0;
  double dm = kmeans_sqr_dist(d, di, mea[0]);
  for (int j = 1; j < k; j++) {
    double dj = kmeans_sqr_dist(d, di, mea[j]);
    if (dj < dm) { dm = dj; m = j; }
  }
  bool change = *cli != m;
  *cli = m;
  return change;
}

// recalcule les moyennes
void kmeans_update(int k, int d, double **data, int n, double **mea, int *cl) {
  int *cnt = calloc(k, sizeof(int));
  double **sum = calloc(k, sizeof(double*));
  for (int j = 0; j < k; j++) {
    cnt[j] = 0;
    sum[j] = calloc(d, sizeof(double));
  }
  for (int i = 0; i < n; i++) {
    int j = cl[i];
    cnt[j]++;
    for (int l = 0; l < d; l++)
      sum[j][l] += data[i][l];
  }
  for (int j = 0; j < k; j++) {
    if (cnt[j] > 0) // si la classe j est vide, on laisse la moyenne inchangée
      for (int l = 0; l < d; l++)
        mea[j][l] = sum[j][l] / cnt[j];
  }
  free(cnt);
  for (int j = 0; j < k; j++) free(sum[j]);
  free(sum);
}

double **kmeans(int k, int d, double **data, int n, int *cl) {
  assert(n >= k);
  double **mea = calloc(k, sizeof(double*));
  for (int j = 0; j < k; j++)
    mea[j] = calloc(d, sizeof(double));
  // initialisation avec k points aléatoires
  kmeans_sampling(k, d, data, n, mea);
  int steps = 0;
  while (true) {
    steps++;
    // classifier
    bool change = false;
    for (int i = 0; i < n; i++)
      change |= kmeans_classify(k, d, data[i], mea, &cl[i]);
    if (!change) break;
    // mettre à jour les moyennes
    kmeans_update(k, d, data, n, mea, cl);
  }
  printf("%d steps\n", steps);
  return mea;
}

